/*! \mainpage Compute Mesh Structural Distortion Metric

According to: <i>Perceptually driven 3D distance metrics with application to watermarking</i>, 
Lavou&eacute;, G., Drelie Gelasca, E., Dupont, F., Baskurt, A., Ebrahimi, T., 
in Proceedings of SPIE Applications of Digital Image Processing XXIX, August 2006.

This is a mex package to use the MSDM 3D distortion metric.

\author Guillaume Lavou&eacute;, Xavier Rolland-Nevi&egrave;re

\section Use

The source code allows to generates the project to create the mex library for the MSDM. Although it also contains the already-built MSDM library, you need 
the following to fully (re)-use and recompile the provided code:
-# MATLAB installation: this may be obvious, but MATLAB is not cheap. You might be able to get around using Octave, but you will need
to make appropriate changes to the link and includes folders in *CMakeLists.txt*. If you don't want to use MATLAB, you can hack and slash into the cpp code
and remove the gateway function to create a traditional dll.
-# Visual Studio 2010 installation (MSVC 2010). 
If you are using another platform than Windows, you will need to modify the *CMakeLists.txt* file.
If you are using another version of Visual Studio, things should still work, but recall that CGAL has its own requirements.
-# Eigen library (http://eigen.tuxfamily.org/index.php?title=Main_Page), which is a very useful and elegant set of headers.
-# CGAL installation: download and install CGAL from https://www.cgal.org/. It is an open-source project and the licence for non-commercial
use is not restrictive. Two important things when you install CGAL:
	-# CGAL has to be installed **for the same architecture as MATLAB**. If you have a x64 version of MATLAB, you will need to have a 
compatible x64 installation of the libraries below.
	-# CGAL relies on third-party libraries, such as QT and QGLViewer for visualization. In my latest tests, this part of the install
crashed for my x64 architecture. As you won't need them, since you *only need the core CGAL libraries*, you can skip their installations
and the compilation of demos and examples in CGAL.
-# Boost installation, which is required to build the CGAL libraries. Download and install from http://www.boost.org/.
-# Cmake installation (http://www.cmake.org/), used to set all the parameters of the MSVC solution.
-# Doxygen installation (http://www.stack.nl/~dimitri/doxygen/) to generate the present documentation. 
If not present, this generation is bypassed.

Once everything is installed, you can use the CMake file provided to generate a MSVC 2010 solution with a main CGAL_MSDM project.
Upon compilation, there is a post-build command to move the MATLAB mex library into the 'deploy' folder.

In this example, *mCGAL_MSDM.m* is the MATLAB wrapper function to call the simplification mex library. 
While mex libraries may be directly called from MATLAB, this encapsulation is useful to write 
help documentation accessible within MATLAB, perform some parsing operations with a high-level language, and perform 
checks.

MATLAB does not provide a simple way for mex function to respond to Ctrl-C interrupt, so to kill the routine, you have to kill MATLAB.

*/

#ifndef CGAL_MSDM_H
#define CGAL_MSDM_H

#include "Header.h"
#include "MatlabPolyhedron.h"
#include "normal_cycle.h"
#include "Distance3D.h"

/*! \file CGAL_MSDM.h
\brief Declaration of the the gateway functions between MATLAB and C++
\author Guillaume Lavou&eacute;, Xavier Rolland-Nevi&egrave;re*/

/*! \brief Initialize the various options of the MSDM estimation
\param options options of the MSDM whose values are initialized to the defaults

Default values are indicated in the documentation on MSDM_options.
*/
void init_MSDM_options(MSDM_options& options);

/*! \brief Parse MATLAB structure options and accordingly modify the MSDM option structure previously default intialized
\param MATLAB_MSDM_options MSDM option structure (MATLAB style) that may be only partially filled
\param options MSDM option structure (C style)
*/
void parse_MSDM_options(const mxArray* MATLAB_MSDM_options, MSDM_options& options);

/*!	\brief Standard gateway between MATLAB and C++
\param *prhs[] Pointer to right-hand side array of matrices in MATLAB (argument to the mex library)
\param nrhs Size of the prhs[] array
\param *plhs[] Pointer to left-hand side array of matrices in MATLAB (output to the mex library)
\param nlhs Size of the plhs[] array
\author Xavier Rolland-Nevi&egrave;re

prhs[0] is a 3 by number-of-vertices MATLAB matrix of vertex locations and prhs[1] is a 3 by number-of-facet MATLAB matrix 
linking vertices to form triangle facets. prhs[2] and prhs[3] are the corresponding matrices for the distorted version of
the mesh. prhs[4] is a scalar indicating the radius of the search. prhs[5] is optional and 
represents a MATLAB matrix with an option structure. Its field are detailed in the MSDM_option structure.

<b>Prerequisite: the MSDM only works for polyhedron sharing the same connectivity, i.e. prhs[1] and prhs[3] shoud be
identical. </b>
*/

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]);

/*!	\brief Main computation procedure to estimate the MSDM between 2 meshes
\param p1 Original polyhedron
\param p2 Distorted polyhedron
\param options Parameter settings
\param [out] local_MSDM vector of local MSDM value estimated at every vertex location

See the MSDM paper for more details on the steps of the procedure.
*/
double compute_MSDM(Polyhedron& p1, Polyhedron& p2, const MSDM_options& options, std::vector<double>& local_MSDM);

#endif