/*\file Distance3D.cpp
\brief Implementation of the MSDM computation from the vertex curvature analysis

\author Guillaume Lavou&eacute;, Xavier Rolland-Nevi&egrave;re
*/

#include "Distance3D.h"

Distance3D::Distance3D(Polyhedron *_PolyOriginal, Polyhedron *_PolyDegrad, const MSDM_options& options)
{
	m_PolyOriginal = _PolyOriginal;
	m_PolyDegrad = _PolyDegrad;
	mOptions = options;
}

mwSize Distance3D::buildPairwiseEstimation()
{	
	double maxdim = m_PolyOriginal->getMaxDim();
	mp_radius = maxdim * mOptions.radius;
	mwSize failures = 0;
	Polyhedron::Vertex_iterator	pVertexDeg = m_PolyDegrad->vertices_begin();
	Polyhedron::Vertex_iterator	pVertex	= m_PolyOriginal->vertices_begin();
	mp_wi_alpha = 1 / (0.008 * maxdim * std::sqrt(2 * pi));
	mp_wi_beta = 1 / (2 * 0.008 * 0.008 * maxdim * maxdim);
	if (!mOptions.multithreading)
	{
		CGAL_For_all(pVertex, m_PolyOriginal->vertices_end())
		{		
			std::vector<double> TabDistance, TabDistanceDeg;
			std::vector<Point> TabPoint, TabPointDeg;
			try
			{
				vertex_curvature_average_variance(m_PolyOriginal, pVertex, TabDistance, TabPoint);			
				vertex_curvature_average_variance(m_PolyDegrad, pVertexDeg, TabDistanceDeg, TabPointDeg);
				vertex_curvature_covariance(pVertex, pVertexDeg, TabDistance, TabPoint, TabDistanceDeg, TabPointDeg);
			}
			catch(...){failures++;}
			pVertexDeg++;
		}
	}
	else
	{
		int nb_vertices = (int) m_PolyOriginal->size_of_vertices();
		std::vector<Polyhedron::Vertex_handle> verticesI, verticesD;
		verticesI.reserve(nb_vertices);
		verticesD.reserve(nb_vertices);
		CGAL_For_all(pVertex, m_PolyOriginal->vertices_end())
		{
			verticesI.push_back(pVertex);
			verticesD.push_back(pVertexDeg);
			pVertexDeg++;
		}
#pragma omp parallel for num_threads(mOptions.nb_threads)
		for (int i = 0 ; i < nb_vertices ; i++)
		{
			std::vector<double> TabDistance, TabDistanceDeg;
			std::vector<Point> TabPoint, TabPointDeg;	
			try
			{
				vertex_curvature_average_variance(m_PolyOriginal, verticesI[i], TabDistance, TabPoint);			
				vertex_curvature_average_variance(m_PolyDegrad, verticesD[i], TabDistanceDeg, TabPointDeg);
				vertex_curvature_covariance(verticesI[i], verticesD[i], TabDistance, TabPoint, TabDistanceDeg, TabPointDeg);
			}
			catch(...)
			{
#pragma omp atomic
				failures++;
			}
		}
	}
	return failures;
}

double Distance3D::updateMSDM(std::vector<double>& lMSDM) const
{
	lMSDM.clear();
	lMSDM.reserve(m_PolyOriginal->size_of_vertices());
	double SommeDistM = 0;
	Polyhedron::Vertex_const_iterator	pVertexDeg = m_PolyDegrad->vertices_begin();
	Polyhedron::Vertex_const_iterator	pVertex	= m_PolyOriginal->vertices_begin();
	Polyhedron::Vertex_const_iterator	pVertexE = m_PolyOriginal->vertices_end();
	CGAL_For_all(pVertex, pVertexE)
	{				
		double MoyX = pVertex->get_courbureMoyenne();
		double MoyY = pVertexDeg->get_courbureMoyenne();
		double SigX = pVertex->get_courbureVariance();
		double SigY = pVertexDeg->get_courbureVariance();
		double SigXY = pVertex->get_courbureCoVariance();	
		double fact1 = std::abs(MoyX - MoyY) / (std::max(MoyX, MoyY) + 1); //NB: '+1' not in original paper
		double fact2 = std::abs(SigX - SigY) / (std::max(SigX, SigY) + 1);
		double fact3 = std::abs(SigX * SigY - SigXY) / (SigX * SigY + 1);	
		double angleRadM = mOptions.alpha * std::pow(fact1, mOptions.Mpower) + 
			mOptions.beta * std::pow(fact2, mOptions.Mpower) + 
			mOptions.gamma * std::pow(fact3, mOptions.Mpower);
		lMSDM.push_back(std::pow(angleRadM, 1 / mOptions.Mpower));
		SommeDistM += angleRadM;
		pVertexDeg++;
	}
	return std::pow(SommeDistM / (double) m_PolyOriginal->size_of_vertices(), 1 / mOptions.Mpower);
}


inline double Distance3D::weighted_covariance(const std::vector<double>& TabD, const std::vector<double>& TabD2, 
	int ecart, const std::vector<Point>& TabP, const std::vector<Point>& TabP2, 
	const Point& O, double moyenne, double moyenneDeg)
{
	double somme1 = 0;
	double sommewi = 0;
	double x1 = 0;
	double x2 = 0;
	double buff = 0;
	double maxBuff = 0;
	for(int i = 0 ; i < (int)TabD.size() ; i++)
	{
		int IndMin = (i < 2 * ecart) ? 0 : (i - 2 * ecart);
		int IndMax = (i + 2 * ecart > TabP2.size()) ? (int) TabP2.size() : i + 2 * ecart;			
		x1 = TabD[i];
		x2 = TabD2[IndMin];
		Vector_3 VBuff = TabP2[IndMin] - TabP[i];
		maxBuff = VBuff.squared_length();
		for(int j = IndMin ; j < IndMax ; j++)
		{			
			VBuff = TabP2[j] - TabP[i];			
			buff = VBuff.squared_length();
			if (buff < maxBuff)
			{
				maxBuff = buff;
				x2 = TabD2[j];
			}
		}
		Vector_3 DistancePt = TabP[i] - O;	
		double wi = mp_wi_alpha * std::exp(-(DistancePt.squared_length()) * mp_wi_beta);		
		sommewi += wi;			
		somme1 += wi * (x1 - moyenne) * (x2 - moyenneDeg);
	}
	return (sommewi != 0) ? somme1 / sommewi : 0;
}

void Distance3D::vertex_curvature_covariance(Polyhedron::Vertex_handle pVertex, Polyhedron::Vertex_handle pVertexDeg, const std::vector<double>& TabDistance, 
	const std::vector<Point>& TabPoint, const std::vector<double>& TabDistanceDeg, const std::vector<Point>& TabPointDeg)
{
	int taille1 = (int) TabDistance.size();
	int taille2 = (int) TabDistanceDeg.size();
	int ecart = std::max(taille1, taille2) - std::min(taille1, taille2);
	CGAL_precondition(taille1 == TabPoint.size());
	CGAL_precondition(taille2 == TabPointDeg.size());
	if (taille1 == 0 || taille2 == 0)
	{
		pVertexDeg->set_courbureCoVariance(0);
		pVertex->set_courbureCoVariance(0);
	}
	double moyenne = pVertex->get_courbureMoyenne();
	double moyenneDeg = pVertexDeg->get_courbureMoyenne();
	double c1 = weighted_covariance(TabDistance, TabDistanceDeg, ecart, TabPoint, TabPointDeg, pVertexDeg->point(), moyenne, moyenneDeg);
	double c2 = weighted_covariance(TabDistanceDeg, TabDistance, ecart, TabPointDeg, TabPoint, pVertexDeg->point(), moyenneDeg, moyenne);
	double cov = (c1 + c2) / 2;
	pVertexDeg->set_courbureCoVariance(cov);
	pVertex->set_courbureCoVariance(cov);
}

void Distance3D::vertex_curvature_average_variance(Polyhedron* PolyUsed, Polyhedron::Vertex_handle pVertex, 
	std::vector<double>& TabDistance, std::vector<Point>& TabPoint)
{
	std::set<Polyhedron::Vertex_const_handle> vertices;
	const Point& O = pVertex->point();
	std::stack<Polyhedron::Vertex_const_handle> S;
	S.push(pVertex);	
	vertices.insert(pVertex);
	int NbSommetInSphere = 0;
	while(!S.empty())
	{
		Polyhedron::Vertex_const_handle v = S.top();
		S.pop();
		const Point& P = v->point();
		Polyhedron::Halfedge_around_vertex_const_circulator h = v->vertex_begin();
		Polyhedron::Halfedge_around_vertex_const_circulator pHalfedgeStart = h;
		CGAL_For_all(h, pHalfedgeStart)
		{
			const Point& p1 = h->vertex()->point();
			const Point& p2 = h->opposite()->vertex()->point();
			Vector_3 V = (p2 - p1);
			if(v == pVertex || V * (P - O) > 0.0) 
			{
				double len_old = std::sqrt(V.squared_length());
				bool isect = sphere_clip_vector(O, mp_radius, P, V);
				double len_edge = std::sqrt(V.squared_length());
				NbSommetInSphere++;
				double DistancePondere;
				Point PPondere;
				if(len_old != 0)
				{
					DistancePondere = (1 - len_edge / len_old) * h->vertex()->get_courbureMSDM() + 
						len_edge / len_old * h->opposite()->vertex()->get_courbureMSDM();
					PPondere = p1 + V;
				}
				else
				{
					DistancePondere = h->opposite()->vertex()->get_kmax();
					PPondere = p2;
				}
				TabDistance.push_back(DistancePondere);
				TabPoint.push_back(PPondere);
				if(!isect) 
				{
					Polyhedron::Vertex_const_iterator w = h->opposite()->vertex();
					if(vertices.find(w) == vertices.end())
					{
						vertices.insert(w);
						S.push(w);
					}
				}
			}
		}			
	}
	std::vector<double> tab_wi;
	double SommeDistance = 0;
	double SommeWi = 0;
	for(int i = 0 ; i < NbSommetInSphere ; i++)
	{
		Vector_3 DistancePt = TabPoint[i] - pVertex->point();
		double wi = mp_wi_alpha * std::exp(-DistancePt.squared_length() * mp_wi_beta);
		tab_wi.push_back(wi);
		SommeWi += wi;
		SommeDistance += TabDistance[i] * wi;
	}
	pVertex->set_courbureMoyenne(SommeDistance / (double) SommeWi);
	double variance = 0;
	for(int i = 0 ; i < NbSommetInSphere ; i++)
		variance += tab_wi[i] * std::pow(TabDistance[i] - pVertex->get_courbureMoyenne(), 2);			
	pVertex->set_courbureVariance(std::sqrt(variance / SommeWi));
}