/*! \file enriched_polyhedron.h
	\brief Enriched CGAL Polyhedron for the MSDM estimation
	\author Guillaume Lavou&eacute;, Xavier Rolland-Nevi&egrave;re
*/

#ifndef _POLYGON_MESH_
#define _POLYGON_MESH_

/*!	\brief Enriched facet for CGAL polyhedron adding normal and tag information

	See the CGAL documentation for information on templates
*/
template <class Refs, class T>
class Enriched_facet : public CGAL::HalfedgeDS_face_base<Refs, T>
{	
	std::size_t m_tag; 
	Vector_3 m_normal;
	double m_area;
public:
	Enriched_facet(){}
	
	/*! \brief Return facet normal vector
		\return Const reference to the facet normal

		Normal vector are oriented outward for consistently oriented polyhedron.
	*/
	const Vector_3& normal() const { return m_normal; }

	/*! \brief Getter to facet tag
		\return Value of the tag */
	std::size_t get_tag(){return m_tag;}

	/*! \brief Setter to facet tag
		\param tag Value of the tag	*/
	void set_tag(std::size_t tag){m_tag = tag;}

	/*! \brief Getter to surface area of the triangle facet*/
	double get_area() const{return m_area;}

	/*! \brief Update internal data on the facet

		Call this function to initialize the internal information on the facet: the surface area and normal
	*/
	void update()
	{
		Point& p0 = this->halfedge()->vertex()->point();
		Point& p1 = this->halfedge()->next()->vertex()->point();
		Point& p2 = this->halfedge()->next()->next()->vertex()->point();
		m_area = std::sqrt(Triangle(p0, p1, p2).squared_area());
		m_normal = normalize(K::Plane_3(p0, p1, p2).orthogonal_vector());
	}
};

/*! \brief Enriched halfedge for CGAL polyhedron adding tag information

	See the CGAL documentation for information on templates

*/
template <class Refs, class Tprev, class Tvertex, class Tface>
class Enriched_halfedge : public CGAL::HalfedgeDS_halfedge_base<Refs, Tprev, Tvertex, Tface>
{
private:
	std::size_t m_tag; 
public:

	/*! \brief Getter to halfedge tag
		\return Value of the tag */
	std::size_t get_tag(){return m_tag;}

	/*! \brief Setter to halfedge tag
		\param tag Value of the tag	*/
	void set_tag(std::size_t tag){m_tag = tag;}
};

/*! \brief Enriched vertex with a tag and maximum principale curvature.

	See the CGAL documentation for information on templates.
*/
template <class Refs, class T>
class Enriched_vertex : public CGAL::HalfedgeDS_vertex_base<Refs, T, Point>
{
	std::size_t m_tag;
	double mp_kmax, mp_kmin, courbureMSDM;
	double courbureMoyenne, courbureVariance, courbureCoVariance;
public:
	
	/*! \brief Default constructor*/
	Enriched_vertex()  {}

	/*! \brief Constructor of vertex at a given location
		\param pt Location of the vertex
	*/
	Enriched_vertex(const Point& pt) : CGAL::HalfedgeDS_vertex_base<Refs, T, Point>(pt){}

	/*! \brief Getter to vertex tag
		\return Value of the tag */
	inline std::size_t get_tag() const {return m_tag;}

	/*! \brief Setter to vertex tag
		\param tag Value of the tag	*/
	inline void set_tag(std::size_t tag){m_tag = tag;}

	/*! \brief Getter to maximum curvature
		\return Value of the maximum curvature. */
	inline double get_kmax() const {return mp_kmax;}

	/*! \brief Setter to maximum curvature
		\param kmax Value of the maximum curvature. */
	inline void set_kmax(double kmax){mp_kmax = kmax;}

	/*! \brief Getter to minimum curvature
		\return Value of the minimum curvature. */
	inline double get_kmin() const {return mp_kmin;}

	/*! \brief Setter to minimum curvature
		\param kmin Value of the minimum curvature. */
	inline void set_kmin(double kmin){mp_kmin = kmin;}

	/*! \brief Getter to MSDM scaled mean curvature
		\return Value of the scaled mean curvature. */
	inline double get_courbureMSDM() const {return courbureMSDM;}

	/*! \brief Setter to MSDM scaled mean curvature
		\param value Value of the scaled mean curvature. */
	inline void set_courbureMSDM(double value){courbureMSDM = value;}	

	
	/*! \brief Getter to the local weighted average of maximum curvatures.
		\return courbureMoyenne
	
		Not to be confused with mean curvature k_mean. This is the weighted average used  by the 
		MSDM.
	*/
	inline double get_courbureMoyenne() const {return courbureMoyenne;}

	
	/*! \brief Setter to the local weighted average of maximum curvatures.
		\param value Value of courbureMoyenne.
		
		Not to be confused with mean curvature k_mean. This is the weighted average used  by the 
		MSDM.
	*/
	inline void set_courbureMoyenne(double value){courbureMoyenne = value;}

	/*! \brief Getter to the local weighted variance of maximum curvatures.
		\return courbureVariance.*/
	inline double get_courbureVariance() const {return courbureVariance;}

	/*! \brief Setter to the local weighted variance of maximum curvatures.
		\param value Value of courbureVariance.*/
	inline void set_courbureVariance(double value){courbureVariance = value;}

	/*! \brief Getter to the local weighted covariance of maximum curvatures.
		\return courbureCoVariance.*/
	inline double get_courbureCoVariance() const {return courbureCoVariance;}

		/*! \brief Setter to the local weighted covariance of maximum curvatures.
		\param value Value of courbureCoVariance.*/
	inline void set_courbureCoVariance(double value){courbureCoVariance = value;}
};

/*! \brief A redefined items class for the Polyhedron_3 

	A refined vertex class that contains a member for the normal vector and a refined facet with a normal vector instead of the 
	plane equation (this is an alternative solution instead of using Polyhedron_traits_with_normals_3).

*/
struct Enriched_items : public CGAL::Polyhedron_items_3
{
	/*! \brief Wrapper for the vertex used in the custom polyhedron*/
	template <class Refs, class Traits>
	struct Vertex_wrapper{ typedef Enriched_vertex<Refs, CGAL::Tag_true> Vertex;};

	/*! \brief Wrapper for the face used in the custom polyhedron*/
	template <class Refs, class Traits>
	struct Face_wrapper{typedef Enriched_facet<Refs, CGAL::Tag_true> Face;};

	/*! \brief Wrapper for the halfedge used in the custom polyhedron*/
	template <class Refs, class Traits>
	struct Halfedge_wrapper{typedef Enriched_halfedge<Refs, CGAL::Tag_true, CGAL::Tag_true, CGAL::Tag_true> Halfedge;};
};


/*! \brief Polyhedron class for the MSDM computation
	
	Enriched Polyhedron with enriched items containing facet normals and surfaces, and a bounding box.
*/
template <class kernel, class items>
class Enriched_polyhedron : public CGAL::Polyhedron_3<kernel,items>
{
private :
	double maxDim;
	CGAL::Bbox_3 m_bbox;
public :

	/*! \brief Build information on facet, vertices, surface and bounding box.

		The building process starts by initializing facet normals, tags and surface information.
		The bounding box is then set.

		<b>This function should be called prior to any computation using the polyhedron.</b>
	*/
	void construct()
	{
		std::size_t facet_id = 0;
		Polyhedron::Facet_iterator fit = facets_begin();
		CGAL_For_all (fit, facets_end())
		{		
			fit->set_tag(facet_id++);
			fit->update();
		}
		Polyhedron::Vertex_iterator vi = vertices_begin();
		std::size_t vertex_id = 0;
		m_bbox = CGAL::Bbox_3(points_begin()->bbox());
		CGAL_For_all(vi, vertices_end())
		{
			vi->set_tag(vertex_id++);
			m_bbox += vi->point().bbox();
		}
		double X = m_bbox.xmax() - m_bbox.xmin();
		double Y = m_bbox.ymax() - m_bbox.ymin();
		double Z = m_bbox.zmax() - m_bbox.zmin();
		maxDim = std::max(std::max(X, Y), Z);
	}

	/*!	\brief Return the maximum of the bounding box sides.*/
	double getMaxDim() const {return maxDim;}
};


#endif